@@ -1,9 +1,11 @@
 #
 # This file is part of pyasn1 software.
 #
-# Copyright (c) 2005-2018, Ilya Etingof <etingof@gmail.com>
+# Copyright (c) 2005-2019, Ilya Etingof <etingof@gmail.com>
 # License: http://snmplabs.com/pyasn1/license.html
 #
+import sys
+
 from pyasn1 import debug
 from pyasn1 import error
 from pyasn1.codec.ber import eoo
@@ -17,6 +19,8 @@ from pyasn1.type import useful
 
 __all__ = ['encode']
 
+LOG = debug.registerLoggee(__name__, flags=debug.DEBUG_ENCODER)
+
 
 class AbstractItemEncoder(object):
     supportIndefLenMode = True
@@ -31,29 +35,39 @@ class AbstractItemEncoder(object):
         encodedTag = tagClass | tagFormat
         if isConstructed:
             encodedTag |= tag.tagFormatConstructed
+
         if tagId < 31:
             return encodedTag | tagId,
+
         else:
             substrate = tagId & 0x7f,
+
             tagId >>= 7
+
             while tagId:
                 substrate = (0x80 | (tagId & 0x7f),) + substrate
                 tagId >>= 7
+
             return (encodedTag | 0x1F,) + substrate
 
     def encodeLength(self, length, defMode):
         if not defMode and self.supportIndefLenMode:
             return (0x80,)
+
         if length < 0x80:
             return length,
+
         else:
             substrate = ()
             while length:
                 substrate = (length & 0xff,) + substrate
                 length >>= 8
+
             substrateLen = len(substrate)
+
             if substrateLen > 126:
                 raise error.PyAsn1Error('Length octets overflow (%d)' % substrateLen)
+
             return (0x80 | substrateLen,) + substrate
 
     def encodeValue(self, value, asn1Spec, encodeFun, **options):
@@ -75,26 +89,51 @@ class AbstractItemEncoder(object):
 
         defMode = options.get('defMode', True)
 
+        substrate = null
+
         for idx, singleTag in enumerate(tagSet.superTags):
 
             defModeOverride = defMode
 
             # base tag?
             if not idx:
-                substrate, isConstructed, isOctets = self.encodeValue(
-                    value, asn1Spec, encodeFun, **options
-                )
+                try:
+                    substrate, isConstructed, isOctets = self.encodeValue(
+                        value, asn1Spec, encodeFun, **options
+                    )
+
+                except error.PyAsn1Error:
+                    exc = sys.exc_info()
+                    raise error.PyAsn1Error(
+                        'Error encoding %r: %s' % (value, exc[1]))
+
+                if LOG:
+                    LOG('encoded %svalue %s into %s' % (
+                        isConstructed and 'constructed ' or '', value, substrate
+                    ))
 
                 if not substrate and isConstructed and options.get('ifNotEmpty', False):
                     return substrate
 
-                # primitive form implies definite mode
                 if not isConstructed:
                     defModeOverride = True
 
+                    if LOG:
+                        LOG('overridden encoding mode into definitive for primitive type')
+
             header = self.encodeTag(singleTag, isConstructed)
+
+            if LOG:
+                LOG('encoded %stag %s into %s' % (
+                    isConstructed and 'constructed ' or '',
+                    singleTag, debug.hexdump(ints2octs(header))))
+
             header += self.encodeLength(len(substrate), defModeOverride)
 
+            if LOG:
+                LOG('encoded %s octets (tag + payload) into %s' % (
+                    len(substrate), debug.hexdump(ints2octs(header))))
+
             if isOctets:
                 substrate = ints2octs(header) + substrate
 
@@ -131,6 +170,11 @@ class IntegerEncoder(AbstractItemEncoder):
 
     def encodeValue(self, value, asn1Spec, encodeFun, **options):
         if value == 0:
+            if LOG:
+                LOG('encoding %spayload for zero INTEGER' % (
+                    self.supportCompactZero and 'no ' or ''
+                ))
+
             # de-facto way to encode zero
             if self.supportCompactZero:
                 return (), False, False
@@ -157,11 +201,15 @@ class BitStringEncoder(AbstractItemEncoder):
             substrate = alignedValue.asOctets()
             return int2oct(len(substrate) * 8 - valueLength) + substrate, False, True
 
+        if LOG:
+            LOG('encoding into up to %s-octet chunks' % maxChunkSize)
+
         baseTag = value.tagSet.baseTag
 
         # strip off explicit tags
         if baseTag:
             tagSet = tag.TagSet(baseTag, baseTag)
+
         else:
             tagSet = tag.TagSet()
 
@@ -195,44 +243,47 @@ class OctetStringEncoder(AbstractItemEncoder):
         if not maxChunkSize or len(substrate) <= maxChunkSize:
             return substrate, False, True
 
-        else:
+        if LOG:
+            LOG('encoding into up to %s-octet chunks' % maxChunkSize)
 
-            # strip off explicit tags for inner chunks
+        # strip off explicit tags for inner chunks
 
-            if asn1Spec is None:
-                baseTag = value.tagSet.baseTag
+        if asn1Spec is None:
+            baseTag = value.tagSet.baseTag
 
-                # strip off explicit tags
-                if baseTag:
-                    tagSet = tag.TagSet(baseTag, baseTag)
-                else:
-                    tagSet = tag.TagSet()
+            # strip off explicit tags
+            if baseTag:
+                tagSet = tag.TagSet(baseTag, baseTag)
 
-                asn1Spec = value.clone(tagSet=tagSet)
+            else:
+                tagSet = tag.TagSet()
 
-            elif not isOctetsType(value):
-                baseTag = asn1Spec.tagSet.baseTag
+            asn1Spec = value.clone(tagSet=tagSet)
 
-                # strip off explicit tags
-                if baseTag:
-                    tagSet = tag.TagSet(baseTag, baseTag)
-                else:
-                    tagSet = tag.TagSet()
+        elif not isOctetsType(value):
+            baseTag = asn1Spec.tagSet.baseTag
 
-                asn1Spec = asn1Spec.clone(tagSet=tagSet)
+            # strip off explicit tags
+            if baseTag:
+                tagSet = tag.TagSet(baseTag, baseTag)
 
-            pos = 0
-            substrate = null
+            else:
+                tagSet = tag.TagSet()
 
-            while True:
-                chunk = value[pos:pos + maxChunkSize]
-                if not chunk:
-                    break
+            asn1Spec = asn1Spec.clone(tagSet=tagSet)
 
-                substrate += encodeFun(chunk, asn1Spec, **options)
-                pos += maxChunkSize
+        pos = 0
+        substrate = null
 
-            return substrate, True, True
+        while True:
+            chunk = value[pos:pos + maxChunkSize]
+            if not chunk:
+                break
+
+            substrate += encodeFun(chunk, asn1Spec, **options)
+            pos += maxChunkSize
+
+        return substrate, True, True
 
 
 class NullEncoder(AbstractItemEncoder):
@@ -268,8 +319,10 @@ class ObjectIdentifierEncoder(AbstractItemEncoder):
                 oid = (second + 80,) + oid[2:]
             else:
                 raise error.PyAsn1Error('Impossible first/second arcs at %s' % (value,))
+
         elif first == 2:
             oid = (second + 80,) + oid[2:]
+
         else:
             raise error.PyAsn1Error('Impossible first/second arcs at %s' % (value,))
 
@@ -280,15 +333,19 @@ class ObjectIdentifierEncoder(AbstractItemEncoder):
             if 0 <= subOid <= 127:
                 # Optimize for the common case
                 octets += (subOid,)
+
             elif subOid > 127:
                 # Pack large Sub-Object IDs
                 res = (subOid & 0x7f,)
                 subOid >>= 7
+
                 while subOid:
                     res = (0x80 | (subOid & 0x7f),) + res
                     subOid >>= 7
+
                 # Add packed Sub-Object ID to resulted Object ID
                 octets += res
+
             else:
                 raise error.PyAsn1Error('Negative OID arc %s at %s' % (subOid, value))
 
@@ -304,12 +361,16 @@ class RealEncoder(AbstractItemEncoder):
         ms, es = 1, 1
         if m < 0:
             ms = -1  # mantissa sign
+
         if e < 0:
-            es = -1  # exponenta sign 
+            es = -1  # exponent sign
+
         m *= ms
+
         if encbase == 8:
             m *= 2 ** (abs(e) % 3 * es)
             e = abs(e) // 3 * es
+
         elif encbase == 16:
             m *= 2 ** (abs(e) % 4 * es)
             e = abs(e) // 4 * es
@@ -320,6 +381,7 @@ class RealEncoder(AbstractItemEncoder):
                 e -= 1
                 continue
             break
+
         return ms, int(m), encbase, e
 
     def _chooseEncBase(self, value):
@@ -327,23 +389,32 @@ class RealEncoder(AbstractItemEncoder):
         encBase = [2, 8, 16]
         if value.binEncBase in encBase:
             return self._dropFloatingPoint(m, value.binEncBase, e)
+
         elif self.binEncBase in encBase:
             return self._dropFloatingPoint(m, self.binEncBase, e)
-        # auto choosing base 2/8/16 
+
+        # auto choosing base 2/8/16
         mantissa = [m, m, m]
-        exponenta = [e, e, e]
+        exponent = [e, e, e]
         sign = 1
         encbase = 2
         e = float('inf')
+
         for i in range(3):
             (sign,
              mantissa[i],
              encBase[i],
-             exponenta[i]) = self._dropFloatingPoint(mantissa[i], encBase[i], exponenta[i])
-            if abs(exponenta[i]) < abs(e) or (abs(exponenta[i]) == abs(e) and mantissa[i] < m):
-                e = exponenta[i]
+             exponent[i]) = self._dropFloatingPoint(mantissa[i], encBase[i], exponent[i])
+
+            if abs(exponent[i]) < abs(e) or (abs(exponent[i]) == abs(e) and mantissa[i] < m):
+                e = exponent[i]
                 m = int(mantissa[i])
                 encbase = encBase[i]
+
+        if LOG:
+            LOG('automatically chosen REAL encoding base %s, sign %s, mantissa %s, '
+                'exponent %s' % (encbase, sign, m, e))
+
         return sign, m, encbase, e
 
     def encodeValue(self, value, asn1Spec, encodeFun, **options):
@@ -352,69 +423,98 @@ class RealEncoder(AbstractItemEncoder):
 
         if value.isPlusInf:
             return (0x40,), False, False
+
         if value.isMinusInf:
             return (0x41,), False, False
+
         m, b, e = value
+
         if not m:
             return null, False, True
+
         if b == 10:
+            if LOG:
+                LOG('encoding REAL into character form')
+
             return str2octs('\x03%dE%s%d' % (m, e == 0 and '+' or '', e)), False, True
+
         elif b == 2:
             fo = 0x80  # binary encoding
             ms, m, encbase, e = self._chooseEncBase(value)
+
             if ms < 0:  # mantissa sign
                 fo |= 0x40  # sign bit
-            # exponenta & mantissa normalization
+
+            # exponent & mantissa normalization
             if encbase == 2:
                 while m & 0x1 == 0:
                     m >>= 1
                     e += 1
+
             elif encbase == 8:
                 while m & 0x7 == 0:
                     m >>= 3
                     e += 1
                 fo |= 0x10
+
             else:  # encbase = 16
                 while m & 0xf == 0:
                     m >>= 4
                     e += 1
                 fo |= 0x20
+
             sf = 0  # scale factor
+
             while m & 0x1 == 0:
                 m >>= 1
                 sf += 1
+
             if sf > 3:
                 raise error.PyAsn1Error('Scale factor overflow')  # bug if raised
+
             fo |= sf << 2
             eo = null
             if e == 0 or e == -1:
                 eo = int2oct(e & 0xff)
+
             else:
                 while e not in (0, -1):
                     eo = int2oct(e & 0xff) + eo
                     e >>= 8
+
                 if e == 0 and eo and oct2int(eo[0]) & 0x80:
                     eo = int2oct(0) + eo
+
                 if e == -1 and eo and not (oct2int(eo[0]) & 0x80):
                     eo = int2oct(0xff) + eo
+
             n = len(eo)
             if n > 0xff:
                 raise error.PyAsn1Error('Real exponent overflow')
+
             if n == 1:
                 pass
+
             elif n == 2:
                 fo |= 1
+
             elif n == 3:
                 fo |= 2
+
             else:
                 fo |= 3
                 eo = int2oct(n & 0xff) + eo
+
             po = null
+
             while m:
                 po = int2oct(m & 0xff) + po
                 m >>= 8
+
             substrate = int2oct(fo) + eo + po
+
             return substrate, False, True
+
         else:
             raise error.PyAsn1Error('Prohibited Real base %s' % b)
 
@@ -428,6 +528,13 @@ class SequenceEncoder(AbstractItemEncoder):
 
         substrate = null
 
+        omitEmptyOptionals = options.get(
+            'omitEmptyOptionals', self.omitEmptyOptionals)
+
+        if LOG:
+            LOG('%sencoding empty OPTIONAL components' % (
+                    omitEmptyOptionals and 'not ' or ''))
+
         if asn1Spec is None:
             # instance of ASN.1 schema
             value.verifySizeSpec()
@@ -439,23 +546,44 @@ class SequenceEncoder(AbstractItemEncoder):
                     namedType = namedTypes[idx]
 
                     if namedType.isOptional and not component.isValue:
-                            continue
+                        if LOG:
+                            LOG('not encoding OPTIONAL component %r' % (namedType,))
+                        continue
 
                     if namedType.isDefaulted and component == namedType.asn1Object:
-                            continue
+                        if LOG:
+                            LOG('not encoding DEFAULT component %r' % (namedType,))
+                        continue
 
-                    if self.omitEmptyOptionals:
+                    if omitEmptyOptionals:
                         options.update(ifNotEmpty=namedType.isOptional)
 
-                chunk = encodeFun(component, asn1Spec, **options)
-
                 # wrap open type blob if needed
                 if namedTypes and namedType.openType:
-                    wrapType = namedType.asn1Object
-                    if wrapType.tagSet and not wrapType.isSameTypeWith(component):
-                        chunk = encodeFun(chunk, wrapType, **options)
 
-                substrate += chunk
+                    wrapType = namedType.asn1Object
+
+                    if wrapType.typeId in (
+                            univ.SetOf.typeId, univ.SequenceOf.typeId):
+
+                        substrate += encodeFun(
+                                component, asn1Spec,
+                                **dict(options, wrapType=wrapType.componentType))
+
+                    else:
+                        chunk = encodeFun(component, asn1Spec, **options)
+
+                        if wrapType.isSameTypeWith(component):
+                            substrate += chunk
+
+                        else:
+                            substrate += encodeFun(chunk, wrapType, **options)
+
+                            if LOG:
+                                LOG('wrapped with wrap type %r' % (wrapType,))
+
+                else:
+                    substrate += encodeFun(component, asn1Spec, **options)
 
         else:
             # bare Python value + ASN.1 schema
@@ -465,43 +593,85 @@ class SequenceEncoder(AbstractItemEncoder):
                     component = value[namedType.name]
 
                 except KeyError:
-                    raise error.PyAsn1Error('Component name "%s" not found in %r' % (namedType.name, value))
+                    raise error.PyAsn1Error('Component name "%s" not found in %r' % (
+                        namedType.name, value))
 
                 if namedType.isOptional and namedType.name not in value:
+                    if LOG:
+                        LOG('not encoding OPTIONAL component %r' % (namedType,))
                     continue
 
                 if namedType.isDefaulted and component == namedType.asn1Object:
+                    if LOG:
+                        LOG('not encoding DEFAULT component %r' % (namedType,))
                     continue
 
-                if self.omitEmptyOptionals:
+                if omitEmptyOptionals:
                     options.update(ifNotEmpty=namedType.isOptional)
 
-                chunk = encodeFun(component, asn1Spec[idx], **options)
+                componentSpec = namedType.asn1Object
 
                 # wrap open type blob if needed
                 if namedType.openType:
-                    wrapType = namedType.asn1Object
-                    if wrapType.tagSet and not wrapType.isSameTypeWith(component):
-                        chunk = encodeFun(chunk, wrapType, **options)
 
-                substrate += chunk
+                    if componentSpec.typeId in (
+                            univ.SetOf.typeId, univ.SequenceOf.typeId):
+
+                        substrate += encodeFun(
+                                component, componentSpec,
+                                **dict(options, wrapType=componentSpec.componentType))
+
+                    else:
+                        chunk = encodeFun(component, componentSpec, **options)
+
+                        if componentSpec.isSameTypeWith(component):
+                            substrate += chunk
+
+                        else:
+                            substrate += encodeFun(chunk, componentSpec, **options)
+
+                            if LOG:
+                                LOG('wrapped with wrap type %r' % (componentSpec,))
+
+                else:
+                    substrate += encodeFun(component, componentSpec, **options)
 
         return substrate, True, True
 
 
 class SequenceOfEncoder(AbstractItemEncoder):
-    def encodeValue(self, value, asn1Spec, encodeFun, **options):
+    def _encodeComponents(self, value, asn1Spec, encodeFun, **options):
+
         if asn1Spec is None:
             value.verifySizeSpec()
+
         else:
             asn1Spec = asn1Spec.componentType
 
-        substrate = null
+        chunks = []
+
+        wrapType = options.pop('wrapType', None)
 
         for idx, component in enumerate(value):
-            substrate += encodeFun(value[idx], asn1Spec, **options)
+            chunk = encodeFun(component, asn1Spec, **options)
 
-        return substrate, True, True
+            if (wrapType is not None and
+                    not wrapType.isSameTypeWith(component)):
+                # wrap encoded value with wrapper container (e.g. ANY)
+                chunk = encodeFun(chunk, wrapType, **options)
+
+                if LOG:
+                    LOG('wrapped with wrap type %r' % (wrapType,))
+
+            chunks.append(chunk)
+
+        return chunks
+
+    def encodeValue(self, value, asn1Spec, encodeFun, **options):
+        chunks = self._encodeComponents(
+            value, asn1Spec, encodeFun, **options)
+
+        return null.join(chunks), True, True
 
 
 class ChoiceEncoder(AbstractItemEncoder):
@@ -620,13 +790,8 @@ class Encoder(object):
             raise error.PyAsn1Error('Value %r is not ASN.1 type instance '
                                     'and "asn1Spec" not given' % (value,))
 
-        if debug.logger & debug.flagEncoder:
-            logger = debug.logger
-        else:
-            logger = None
-
-        if logger:
-            logger('encoder called in %sdef mode, chunk size %s for '
+        if LOG:
+            LOG('encoder called in %sdef mode, chunk size %s for '
                    'type %s, value:\n%s' % (not options.get('defMode', True) and 'in' or '', options.get('maxChunkSize', 0), asn1Spec is None and value.prettyPrintType() or asn1Spec.prettyPrintType(), value))
 
         if self.fixedDefLengthMode is not None:
@@ -639,8 +804,8 @@ class Encoder(object):
         try:
             concreteEncoder = self.__typeMap[typeId]
 
-            if logger:
-                logger('using value codec %s chosen by type ID %s' % (concreteEncoder.__class__.__name__, typeId))
+            if LOG:
+                LOG('using value codec %s chosen by type ID %s' % (concreteEncoder.__class__.__name__, typeId))
 
         except KeyError:
             if asn1Spec is None:
@@ -657,13 +822,13 @@ class Encoder(object):
             except KeyError:
                 raise error.PyAsn1Error('No encoder for %r (%s)' % (value, tagSet))
 
-            if logger:
-                logger('using value codec %s chosen by tagSet %s' % (concreteEncoder.__class__.__name__, tagSet))
+            if LOG:
+                LOG('using value codec %s chosen by tagSet %s' % (concreteEncoder.__class__.__name__, tagSet))
 
         substrate = concreteEncoder.encode(value, asn1Spec, self, **options)
 
-        if logger:
-            logger('codec %s built %s octets of substrate: %s\nencoder completed' % (concreteEncoder, len(substrate), debug.hexdump(substrate)))
+        if LOG:
+            LOG('codec %s built %s octets of substrate: %s\nencoder completed' % (concreteEncoder, len(substrate), debug.hexdump(substrate)))
 
         return substrate
 
@@ -684,7 +849,7 @@ class Encoder(object):
 #:     Optional ASN.1 schema or value object e.g. :py:class:`~pyasn1.type.base.PyAsn1Item` derivative
 #:
 #: defMode: :py:class:`bool`
-#:     If `False`, produces indefinite length encoding
+#:     If :obj:`False`, produces indefinite length encoding
 #:
 #: maxChunkSize: :py:class:`int`
 #:     Maximum chunk size in chunked encoding mode (0 denotes unlimited chunk size)
@@ -696,7 +861,7 @@ class Encoder(object):
 #:
 #: Raises
 #: ------
-#: :py:class:`~pyasn1.error.PyAsn1Error`
+#: ~pyasn1.error.PyAsn1Error
 #:     On encoding errors
 #:
 #: Examples
